/* Go IPP - IPP core protocol implementation in pure Go
 *
 * Copyright (C) 2020 and up by Alexander Pevzner (pzz@apevzner.com)
 * See LICENSE for license terms and conditions
 *
 * Various constants
 */

package goipp

const (
	// ContentType is the HTTP content type for IPP messages
	ContentType = "application/ipp"

	// msgPrintIndent used for indentation by message pretty-printer
	msgPrintIndent = "    "
)
